use std::mem::size_of;

use zerocopy::{AsBytes, FromBytes, Unaligned};

use crate::{Db, FilePointer, Header, PAGE_SIZE, U32};

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct FreeList {
    head: FilePointer,
}

impl FreeList {
    pub fn empty() -> Self {
        Self {
            head: FilePointer::null(),
        }
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct AllocatorState {
    pub general: FreeList,
    pub slabs: FilePointer,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct PoolListHeader {
    next: FilePointer,
    size: U32,
    len: U32,
}

impl PoolListHeader {
    fn capacity(&self) -> u32 {
        (self.size.get() - size_of::<PoolListHeader>() as u32) / size_of::<SizedFreeList>() as u32
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct SizedFreeList {
    element_size: U32,
    head: FreeList,
}

impl AllocatorState {
    pub fn init(&self, db: &mut Db, size: U32) {
        db.write(
            self.slabs,
            PoolListHeader {
                next: FilePointer::null(),
                size,
                len: 0.into(),
            },
        );
    }

    fn slabs_mut<'db>(&self, db: &'db mut Db) -> &'db mut PoolListHeader {
        db.modify(self.slabs)
    }
}
